/*
 Dahdi layer for Quasar-mini E1 interface board driver

 Copyright (C) 2007 Parabel Ltd.
 Author: Mikhail Khelik <mixel.net@gmail.com>

 Copyright (C) 2008-2009 Parabel Ltd.
 Author: Pavel Selivanov <pavel[AT]parabel.ru>

 * Parabel Ltd. reserve the right to distribute project or it's parts
 under the alternative conditions.

 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software Foundation;
 either version 2 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with this program;
 if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/completion.h>
#include <linux/list.h>
#include <linux/types.h>
#include <linux/interrupt.h>

#include "dahdi/kernel.h"

#include "pmc_e1_conf.h"
#include "pmc_e1.h"
#include "e12zap.h"


#define E1PMC_NUM_SPANS 2                              /* to change you have to rewrite interrupt headler */

struct  e12zap_span
{
    struct dahdi_span span;
    struct dahdi_chan _chans[31];			/* omit 0 timeslot */
    struct dahdi_chan *chans[31];                       /* omit 0 timeslot */
    int spannumber;		                        /* span number in card */
    int syncprior;			                /* use or not this span for syncronization, and syncsource priority */
    void * pvt;
    int clear_counters;
    unsigned char ec_chunk1[31][DAHDI_CHUNKSIZE];          /* first EC chunk buffer */
    unsigned char ec_chunk2[31][DAHDI_CHUNKSIZE];          /* second EC chunk buffer */
};

struct e12zap_pmc
{
  volatile unsigned int *writechunk;			/* Double-word aligned write memory */
  volatile unsigned int *readchunk;                     /* Double-word aligned read memory */
  dma_addr_t 	readdma;
  dma_addr_t	writedma;
  struct e12zap_span    *e12zap_pmc_span[E1PMC_NUM_SPANS];
  unsigned char num_spans;		                        /* how many spans this card has */
 // unsigned char qn_span;                                /*For fast check during the interrupt*/
  char registered;
  void * pvt;
  u32 ovfl_counter;                                     /* keep old value of overflow counter, used in e12zap_delivery */
  unsigned int keep;					/* used for checking interrupt misses */
  int intcount;
  int irqmisses;        				//digium defined it as signed...
};

unsigned long pmc_reg_read (struct pmc_e1_controller *tmp_dev,
			unsigned int spanno, unsigned int offset)
{
    return tmp_dev->bar1[spanno * 0x40 + offset];
}

unsigned int pmc_reg_check (struct pmc_e1_controller *tmp_dev,
			unsigned int spanno, unsigned int offset, unsigned int mask)
{
    return pmc_reg_read(tmp_dev, spanno, offset)
	& mask;
}

void pmc_reg_write (struct pmc_e1_controller *tmp_dev,
			unsigned int spanno, unsigned int offset, unsigned int v)
{
    tmp_dev->bar1[spanno * 0x40 + offset] = v;
}


void e12zap_find_master (struct e12zap_pmc * pmc_zap)
{
	int i;
	struct e12zap_span *ts;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;
	int syncsrc, master;

	master = tmp_dev->bar0[0x02] & 0x07;
	syncsrc = pmc_zap->e12zap_pmc_span[master]->span.spanno;
	ts = pmc_zap->e12zap_pmc_span[master];

	if (tmp_dev->bar0[0x02] & (1<<7)) //Free Run
		syncsrc = 0; //Using internal sync source for CT-BUS

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];
		ts->span.syncsrc = syncsrc;
	}
}

void e12zap_check_alarm (struct e12zap_pmc * pmc_zap)
{
	int i;
	struct e12zap_span *ts;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];

		if (pmc_reg_check(tmp_dev, i, FRS_ADDR_OFFSET, FRS_RAI_RX)) {
			ts->span.alarms |= DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		}
		if (pmc_reg_check(tmp_dev, i, FRS_ADDR_OFFSET, FRS_RED)) { //LOF
			ts->span.alarms |= DAHDI_ALARM_RED;
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_RED;
		}
		if (pmc_reg_check(tmp_dev, i, FRS_ADDR_OFFSET, FRS_AIS)) {
			ts->span.alarms |= DAHDI_ALARM_BLUE;
			ts->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			ts->span.alarms &= ~DAHDI_ALARM_BLUE;
		}
		dahdi_alarm_notify (&ts->span);
		//FAS
                ts->span.fascount = (unsigned short)pmc_reg_read (tmp_dev, i, FRS_FRAMEERR_OFFSET);
		// BPV
		ts->span.bpvcount = (unsigned short)pmc_reg_read (tmp_dev, i, FRS_HDB3ERR_OFFSET);
                //CRC4 ++
		ts->span.crc4count = (unsigned short)pmc_reg_read (tmp_dev, i, FRS_CRC4ERR_OFFSET);
                ts->span.irqmisses = pmc_zap->irqmisses;
        }
}

void e12zap_clearstat (struct e12zap_pmc * pmc_zap) {
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;
	struct e12zap_span *ts;
	int i;

	for (i=0; i<pmc_zap->num_spans; ++i) {
		ts = pmc_zap->e12zap_pmc_span[i];

		if (0 == ts->clear_counters) continue; //Already cleared
		--ts->clear_counters;
		if (0 != ts->clear_counters) continue;
		module_dprintk(KERN_DEBUG, "e12zap_clearstat:"
			       "Clearing span %d error counters\n", ts->span.spanno);
		pmc_reg_write (tmp_dev, i, FRS_FRAMEERR_OFFSET, 0);
		pmc_reg_write (tmp_dev, i, FRS_HDB3ERR_OFFSET, 0);
		pmc_reg_write (tmp_dev, i, FRS_CRC4ERR_OFFSET, 0);
		pmc_zap->irqmisses = 0;
	}
}

static int e12zap_maint(struct dahdi_span *span, int cmd)
{
	//Currently ignore all maint commands
	int res = 0;
	switch(cmd) {
	case DAHDI_MAINT_NONE:
		/* loop=off, rloop=off */
	case DAHDI_MAINT_LOCALLOOP:
		/* loop=on, rloop=? */
	case DAHDI_MAINT_REMOTELOOP:
		/* loop=?, rloop=on */
		break;
	case DAHDI_MAINT_LOOPUP:
	case DAHDI_MAINT_LOOPDOWN:
	case DAHDI_MAINT_LOOPSTOP:
                span->mainttimer = span->maintstat = 0;
		module_printk(KERN_WARNING, "e21zap_maint: Use ecfg testing options.\n");
		res = -ENOSYS;
		break;
	default:
		module_printk(KERN_ERR, "e21zap_maint: Unknown maint command: %d\n", cmd);
		res = -ENOTTY;
		break;
	}
	return res;
}

static int e12zap_rbsbits(struct dahdi_chan *chan, int bits)
{
	//trunk for CAS nibles...
	//put CAS nibbles to appropriate position
	struct e12zap_span * e12zap_pmc_span = chan->pvt;
	struct e12zap_pmc * pmc_zap = e12zap_pmc_span->pvt;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;
	unsigned int tmp;

	module_dprintk(KERN_DEBUG, "e12zap_rbsbits:->%x on %s\n", bits, chan->name);
	if (chan->chanpos == 16) return 0;
 

	if(chan->chanpos < 16){
		//put data
		tmp = tmp_dev->bar1[e12zap_pmc_span->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos];
		tmp &= 0xff0f;
		tmp |=  (bits << 4);
		tmp_dev->bar1[e12zap_pmc_span->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos] = tmp;
	}
	else {
		//put data
		tmp = tmp_dev->bar1[e12zap_pmc_span->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos - 16];
		tmp &= 0xfff0;
		tmp |= bits;
		tmp_dev->bar1[e12zap_pmc_span->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos - 16] = tmp;
	}
	return 0;
}

static int e12zap_chanconfig(struct dahdi_chan *chan, int sigtype)
{
	//hmm... in E1 case, just nothing to do...
	return 0;
}

static int e12zap_ioctl(struct dahdi_chan *chan, unsigned int cmd, unsigned long data)
{
	switch(cmd) {
	default:
		return -ENOTTY;
	}
}

static int e12zap_close(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_close\n");
	module_put(THIS_MODULE);
	return 0;
}

static int e12zap_open(struct dahdi_chan *chan)
{
	module_dprintk(KERN_DEBUG, "e12zap_open\n");
	try_module_get(THIS_MODULE);
	return 0;
}

static int e12zap_shutdown(struct dahdi_span *span)
{
	int wasrunning, i, tmp;
	unsigned char working=0;
	struct e12zap_span *p = span->pvt;
	struct e12zap_pmc * pmc_zap = p->pvt;
	struct pmc_e1_controller * ptmc_dev = pmc_zap->pvt;
	wasrunning = span->flags & DAHDI_FLAG_RUNNING;
	span->flags &= ~DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_shutdown\n");
	if (wasrunning){
		for(i=0;i<pmc_zap->num_spans;i++){
			if(pmc_zap->e12zap_pmc_span[i]->span.flags & DAHDI_FLAG_RUNNING)
				working =1;
		}
	}
	if (!working){
		//disable DMA on recv and transmit
		tmp = ptmc_dev->bar0[0x09];
		tmp &= ~(DMA_WENA|DMA_RENA);
		ptmc_dev->bar0[0x09] = tmp;

		//cancel pending interrupts
		ptmc_dev->bar0[0] = 0;

		//disable DMA interrupt, mask
		tmp = ptmc_dev->bar0[0x01];
		tmp &= ~(0x20000000); //DMA
		//tmp &= ~0x40000000; //RTC
		ptmc_dev->bar0[0x01] = tmp;
	}
	return 0;
}

static int e12zap_startup(struct dahdi_span *span)
{
	unsigned int i, tmp, ucTmp1;
	unsigned char ucTmp=0;
	struct e12zap_span *p = span->pvt;
        struct e12zap_pmc *pmc_zap = p->pvt;
	struct pmc_e1_controller * ptmc_dev = pmc_zap->pvt;
       	int alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_startup: ");

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(p->ec_chunk1[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(p->ec_chunk2[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	dprintk("", "Using ");

	//set page 1;
        tmp = ptmc_dev->bar1[p->spannumber * 0x40];

        if(span->lineconfig & DAHDI_CONFIG_HDB3) //set HDB3
	{
		if (tmp & FRC_SHORT_HAUL_EN) {
			ptmc_dev->bar1[p->spannumber * 0x40] |= (FRC_HDB3_RX_EN | FRC_HDB3_TX_EN) ;
		} else {
			ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4; //HDB3, long haul, JA-tx;
			acc_lxt_reg(ptmc_dev, p->spannumber, LR_CR1, LIU_WRITE, &ucTmp);
                        ucTmp = 0;
                        acc_lxt_reg(ptmc_dev, p->spannumber, LR_CR1, LIU_READ, &ucTmp);
                        if (!(ucTmp & LR_CR1_ENCENB)) {
                       		 module_printk(KERN_ERR, "LIU configuration error\n");
                        }
		} 
		dprintk("", "HDB3/");
 	}
        else //set AMI
	{
            	if (tmp & FRC_SHORT_HAUL_EN) {
			ptmc_dev->bar1[p->spannumber * 0x40] &= ~(FRC_HDB3_RX_EN | FRC_HDB3_TX_EN) ;
	 	} else {
			ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_EC1 | LR_CR1_EC4; //AMI, long haul, JA-tx;
			acc_lxt_reg(ptmc_dev, p->spannumber, LR_CR1, LIU_WRITE, &ucTmp);
                        ucTmp = 0;
                        acc_lxt_reg(ptmc_dev, p->spannumber, LR_CR1, LIU_READ, &ucTmp);
                        if (ucTmp & LR_CR1_ENCENB) {
                       		 module_printk(KERN_ERR, "LIU configuration error\n");
                        }
		} 
		dprintk("", "AMI/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CCS)
	{       //set CSS
		ptmc_dev->bar1[p->spannumber * 0x40] &= ~FRC_MFR_TX_EN;
		dprintk("", "CCS/");
	}
	else  //set CAS
	{
		ptmc_dev->bar1[p->spannumber * 0x40] |= FRC_MFR_TX_EN;
		dprintk("", "CAS/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CRC4)
	{
		ptmc_dev->bar1[p->spannumber * 0x40] |= FRC_CRC4_TX_EN;
		dprintk("", "CRC4 ");
	}
	else
	{
		ptmc_dev->bar1[p->spannumber * 0x40] &= ~FRC_CRC4_TX_EN;
		dprintk("", "non-CRC4 ");
	}

        dprintk("", "on span %s, ", span->name);
	if (p->syncprior) //if !0 allow to use for syncronisation.
        {
	    tmp = ptmc_dev->bar0[0x06]; //get bs/ls
	    ucTmp1 = ~(1<<p->spannumber);
	    tmp&= ucTmp1;
	    ptmc_dev->bar0[0x06] = tmp; //restore bs/ls
            dprintk("", "can be the sync. source\n");
	}
	else
	{
	    tmp = ptmc_dev->bar0[0x06]; //get bs/ls
	    ucTmp1 = (1<<p->spannumber);
	    tmp|= ucTmp1;
	    ptmc_dev->bar0[0x06] = tmp; //restore bs/ls
            dprintk("", "can not be the sync. source\n");
	}

	if (!alreadyrunning) {
	    module_dprintk(KERN_DEBUG, "Enable DMA...\n");
	// Only if we're not already going
	//allow DMA on recv and transmit
	    tmp = ptmc_dev->bar0[0x09];
	    tmp |= (DMA_WENA | DMA_RENA);
	    ptmc_dev->bar0[0x09] = tmp;
	    //allow DMA interrupt
	    tmp = ptmc_dev->bar0[0x01];
	    tmp |= 0x20000000; //DMA
	    //tmp |= 0x40000000; //RTC
	    ptmc_dev->bar0[0x01] = tmp;
	    //restart RTC timer
	    //tmp =2047;
	    //ptmc_dev->bar0[0x10] = tmp;
	    span->flags |= DAHDI_FLAG_RUNNING;
	}
	p->clear_counters = 3;
	return 0;
}

static int e12zap_spanconfig(struct dahdi_span *span, struct dahdi_lineconfig *lc)
{
	struct e12zap_span * p;
	p = (struct e12zap_span *)span->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_spanconfig: Configuring span %d, span=%p\n", span->spanno, span);
	span->lineconfig = lc->lineconfig;
	span->txlevel = lc->lbo;
	span->rxlevel = 0;
	p->syncprior = lc->sync; 	//lineconfig.sync contain the priority of the sysnc source 1-highest 1, 0 - not use this span for the sync at all.
	/* If we're already running, then go ahead and apply the changes */
	if (span->flags & DAHDI_FLAG_RUNNING)
	{
		module_dprintk(KERN_DEBUG, "Starting span!\n");
		return e12zap_startup(span);
	}
	module_dprintk(KERN_DEBUG, "Done with spanconfig!\n");
	return 0;
}

void e12zap_proc(struct e12zap_pmc * pmc_zap);

int e12zap_dacs(struct dahdi_chan *dst, struct dahdi_chan *src);

#define DMA_BUFFER_SIZE 4096

int e12zap_register(struct pmc_e1_controller * ptmc_dev)
{
    struct pci_dev *pdev = ptmc_dev->pci_device;
    int tmp, j, x;
    struct e12zap_pmc * pmc_zap;
    ptmc_dev->priv_data = kmalloc(sizeof(struct e12zap_pmc), GFP_KERNEL);
    if (ptmc_dev->priv_data == NULL) return -ENOMEM;
    memset(ptmc_dev->priv_data,0x0,sizeof(struct e12zap_pmc));
    pmc_zap = (struct e12zap_pmc *) ptmc_dev->priv_data;
    pmc_zap->pvt = ptmc_dev;
    pmc_zap->writechunk  = (unsigned int *)pci_alloc_consistent(pdev, 2 * DMA_BUFFER_SIZE, &pmc_zap->writedma); //2 * 4KB buffers
    if (pmc_zap->writechunk == NULL)
    {
	    module_printk(KERN_ERR, "Unable to allocate DMA buffer\n");
	    kfree(ptmc_dev->priv_data);
	    return -ENOMEM;
    }
    //  printk("DMA phys address=%x\n", pmc_zap->writedma);

    pmc_zap->readchunk = pmc_zap->writechunk + (DMA_BUFFER_SIZE / 4); //double-word address;
    pmc_zap->readdma = pmc_zap->writedma + DMA_BUFFER_SIZE; //physical byte address
	
    module_dprintk(KERN_DEBUG, "Addrs: readchunk: 0x%p, writechunk: 0x%p, readdma: 0x%x, writeddma: %x \n",
		   pmc_zap->readchunk, pmc_zap->writechunk, (unsigned int) pmc_zap->readdma, (unsigned int) pmc_zap->writedma);

    pmc_zap->num_spans = ptmc_dev->e1_amount + 1;
    
    /* Initialize Write/Buffers to all blank data */
    memset((void *)pmc_zap->writechunk,0x00, DMA_BUFFER_SIZE);
    memset((void *)pmc_zap->readchunk,0xff, DMA_BUFFER_SIZE);

    pci_set_master(pdev);

    //allocate memory for the spans
    for (tmp = 0; tmp < pmc_zap->num_spans; tmp++)
    {
	    pmc_zap->e12zap_pmc_span[tmp] = kmalloc(sizeof(struct e12zap_span), GFP_KERNEL);
	    if (pmc_zap->e12zap_pmc_span[tmp])
	    {
		    memset(pmc_zap->e12zap_pmc_span[tmp], 0, sizeof(struct e12zap_span));
	    }
	    else
	    {   //free already allocated
		    for (j=(tmp-1);j >=0; j--)
		    {
			    kfree(pmc_zap->e12zap_pmc_span[tmp]);
		    }
		    goto unreg_all;
	    }
    }
    //initialize spans
    for (tmp = 0; tmp < pmc_zap->num_spans; tmp++)
    {
	    pmc_zap->e12zap_pmc_span[tmp]->spannumber=tmp;
	    pmc_zap->e12zap_pmc_span[tmp]->pvt = pmc_zap;
	    pmc_zap->e12zap_pmc_span[tmp]->span.channels = 31;
	    pmc_zap->e12zap_pmc_span[tmp]->span.spantype = "E1";
	    pmc_zap->e12zap_pmc_span[tmp]->span.manufacturer = "Parabel ltd.";
	    sprintf(pmc_zap->e12zap_pmc_span[tmp]->span.name,
		    "Quasarm_%d_%d", ptmc_dev->dev_number,tmp+1);
	    sprintf(pmc_zap->e12zap_pmc_span[tmp]->span.desc,
		    "Parabel Quasar-mini %d Span %d",ptmc_dev->dev_number, tmp+1);
	    snprintf(pmc_zap->e12zap_pmc_span[tmp]->span.location,
		     sizeof(pmc_zap->e12zap_pmc_span[tmp]->span.location) - 1,
		     "PCI Bus %02d Slot %02d", pdev->bus->number,
		     PCI_SLOT(pdev->devfn) + 1);
	    snprintf(pmc_zap->e12zap_pmc_span[tmp]->span.devicetype,
		     sizeof(pmc_zap->e12zap_pmc_span[tmp]->span.devicetype) - 1,
		     "Quasar-mini PCI, hw:0x%X",
		     ptmc_dev->bar0[0x0D]);
	    pmc_zap->e12zap_pmc_span[tmp]->span.chans = pmc_zap->e12zap_pmc_span[tmp]->chans;
	    pmc_zap->e12zap_pmc_span[tmp]->span.flags = DAHDI_FLAG_RBS;
	    pmc_zap->e12zap_pmc_span[tmp]->span.linecompat = DAHDI_CONFIG_HDB3 | DAHDI_CONFIG_CCS | DAHDI_CONFIG_CRC4 | DAHDI_CONFIG_AMI;  //which modes I can work
	    pmc_zap->e12zap_pmc_span[tmp]->span.deflaw = DAHDI_LAW_ALAW;

	    pmc_zap->e12zap_pmc_span[tmp]->span.pvt =(void *) (pmc_zap->e12zap_pmc_span[tmp]);
	    pmc_zap->e12zap_pmc_span[tmp]->span.spanconfig = e12zap_spanconfig;
	    pmc_zap->e12zap_pmc_span[tmp]->span.chanconfig = e12zap_chanconfig;
	    pmc_zap->e12zap_pmc_span[tmp]->span.startup = e12zap_startup;
	    pmc_zap->e12zap_pmc_span[tmp]->span.shutdown = e12zap_shutdown;
	    pmc_zap->e12zap_pmc_span[tmp]->span.rbsbits = e12zap_rbsbits;
	    pmc_zap->e12zap_pmc_span[tmp]->span.maint = e12zap_maint;
	    pmc_zap->e12zap_pmc_span[tmp]->span.open = e12zap_open;
	    pmc_zap->e12zap_pmc_span[tmp]->span.close  = e12zap_close;
	    pmc_zap->e12zap_pmc_span[tmp]->span.ioctl = e12zap_ioctl;
	    pmc_zap->e12zap_pmc_span[tmp]->span.dacs = e12zap_dacs;
	    init_waitqueue_head(&(pmc_zap->e12zap_pmc_span[tmp]->span.maintq));

	    for (x=0; x < pmc_zap->e12zap_pmc_span[tmp]->span.channels; x++)
	    {
                    pmc_zap->e12zap_pmc_span[tmp]->chans[x] = &pmc_zap->e12zap_pmc_span[tmp]->_chans[x];
		    sprintf(pmc_zap->e12zap_pmc_span[tmp]->chans[x]->name, "Quasarm_%d_%d_%d", ptmc_dev->dev_number, tmp + 1, x + 1);
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->sigcap = DAHDI_SIG_EM | DAHDI_SIG_CLEAR | DAHDI_SIG_EM_E1 |
			    DAHDI_SIG_FXSLS | DAHDI_SIG_FXSGS |
			    DAHDI_SIG_FXSKS | DAHDI_SIG_FXOLS |
			    DAHDI_SIG_FXOGS | DAHDI_SIG_FXOKS |
			    DAHDI_SIG_CAS | DAHDI_SIG_SF | DAHDI_SIG_DACS_RBS;
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->pvt = pmc_zap->e12zap_pmc_span[tmp];
		    pmc_zap->e12zap_pmc_span[tmp]->chans[x]->chanpos = x + 1; //timeslot number
	    }
	    //attach to dahdi
	    if (dahdi_register(&pmc_zap->e12zap_pmc_span[tmp]->span, 0))
	    {
		    module_printk(KERN_ERR, "Unable to register span Quasarm_%d_%d with dahdi!\n", ptmc_dev->dev_number, tmp+1);
		    for (j=(tmp-1);j >=0; j--)
		    {
			    module_dprintk(KERN_DEBUG, "Unregister span %d\n", j);
			    dahdi_unregister(&pmc_zap->e12zap_pmc_span[tmp]->span);
		    }
		    //free spans
		    for (j = 0; j < pmc_zap->num_spans; j++)
		    {
			    kfree(pmc_zap->e12zap_pmc_span[j]);
		    }
		    goto unreg_all;
	    }
    }
    //initialize hardware
    tmp = pmc_zap->readdma;
    ptmc_dev->bar0[0x0a] = tmp;
    tmp = pmc_zap->writedma;
    ptmc_dev->bar0[0x0c] = tmp;
    tmp = DMA_IFQ2 | DMA_SIZE0;
  //  tmp = DMA_IFQ2 | DMA_SIZE0;
    ptmc_dev->bar0[0x09] = tmp;

    pmc_zap->registered = 1;
    return 0;
unreg_all:
	module_dprintk(KERN_DEBUG, "e12zap_register: Free resources!\n");
	pci_free_consistent(pdev, 2 * DMA_BUFFER_SIZE, (void *) pmc_zap->writechunk, pmc_zap->writedma);
	kfree(ptmc_dev->priv_data);
	return -ENOMEM;
}
int e12zap_unregister(struct pmc_e1_controller * ptmc_dev)
{
	struct pci_dev *pdev = ptmc_dev->pci_device;
	struct e12zap_pmc * pmc_zap;
	int j, tmp;
	pmc_zap = (struct e12zap_pmc *) ptmc_dev->priv_data;
	if(pmc_zap)
	{
		if (pmc_zap->registered)
		{
			//detach from dahdi
			for(j=0;j<pmc_zap->num_spans;j++)
			{
				dahdi_unregister(&pmc_zap->e12zap_pmc_span[j]->span);
				kfree(pmc_zap->e12zap_pmc_span[j]);
			}
			pci_free_consistent(pdev, 2 * DMA_BUFFER_SIZE, (void *) pmc_zap->writechunk, pmc_zap->writedma);
			kfree(ptmc_dev->priv_data);
			//deinitialize hardware

			//disable DMA on recv and transmit
			tmp = ptmc_dev->bar0[0x09];
			tmp &= ~(DMA_WENA|DMA_RENA);
			ptmc_dev->bar0[0x09] = tmp;

			//disable DMA interrupt, mask
			tmp = ptmc_dev->bar0[0x01];
			tmp &= ~(0x20000000);
			ptmc_dev->bar0[0x01] = tmp;
		}
	}
    return 0;
}

void e12zap_proc(struct e12zap_pmc * pmc_zap)
{
    int j1,i1;
    for(j1=0; j1 < pmc_zap->num_spans; j1++)
    {
	if( pmc_zap->e12zap_pmc_span[j1]->span.flags &  DAHDI_FLAG_RUNNING)
	{
	    /*Echo cancelation*/
		for(i1=0; i1 < pmc_zap->e12zap_pmc_span[j1]->span.channels; i1++)
		{
			dahdi_ec_chunk(pmc_zap->e12zap_pmc_span[j1]->span.chans[i1],
				    pmc_zap->e12zap_pmc_span[j1]->span.chans[i1]->readchunk,
				    pmc_zap->e12zap_pmc_span[j1]->ec_chunk2[i1]);
			memcpy(pmc_zap->e12zap_pmc_span[j1]->ec_chunk2[i1], pmc_zap->e12zap_pmc_span[j1]->ec_chunk1[i1],
			       DAHDI_CHUNKSIZE);
			memcpy(pmc_zap->e12zap_pmc_span[j1]->ec_chunk1[i1], pmc_zap->e12zap_pmc_span[j1]->span.chans[i1]->writechunk,
			       DAHDI_CHUNKSIZE);
		}
		dahdi_receive(&pmc_zap->e12zap_pmc_span[j1]->span);
	}
    }
    for(j1=0; j1 < pmc_zap->num_spans; j1++)
    {
	    if( pmc_zap->e12zap_pmc_span[j1]->span.flags &  DAHDI_FLAG_RUNNING)
	    {
		    dahdi_transmit(&pmc_zap->e12zap_pmc_span[j1]->span);
	    }
    }
}
/*

RX and TX buffers - 4096 bytes wide.
While DMA controller reads or writes
data from/to buffer addres is 
increasing from 0 to 4095.
 

 0  --------------------------
    I		0	     I
512 --------------------------
    I		1	     I
1024--------------------------
    I		2	     I
1536--------------------------
    I	        3            I
2048--------------------------
    I	    	4	     I
2560--------------------------
    I		5	     I
3072--------------------------
    I		6	     I
3584--------------------------
    I		7	     I
4095--------------------------


*/


#define SET_POINTERS(read_offset, write_offset)  \
	   for(i=0;i<31;i++) { 		\
                pmc_zap->e12zap_pmc_span[0]->chans[i]->readchunk =  (unsigned char *) ((unsigned char *) pmc_zap->readchunk + read_offset + 8 /* skip zero ts*/ +  i*8); /*from the line*/ \
                pmc_zap->e12zap_pmc_span[1]->chans[i]->readchunk =  (unsigned char *) ((unsigned char *) pmc_zap->readchunk + read_offset + 8 + 256 /* dma line 1 offset */ + i*8);        \
                pmc_zap->e12zap_pmc_span[0]->chans[i]->writechunk = (unsigned char *) ((unsigned char *) pmc_zap->writechunk + write_offset + 8 + i*8); /* to the network */              \
                pmc_zap->e12zap_pmc_span[1]->chans[i]->writechunk = (unsigned char *) ((unsigned char *) pmc_zap->writechunk + write_offset + 8 + 256 + i*8);                             \
            }

int e12zap_delivery(struct pmc_e1_controller * ptmc_dev)
{

    struct e12zap_pmc * pmc_zap = ptmc_dev->priv_data;
    unsigned int keep, i, tmp1, j1;
    //unsigned char cur_part;
    /*   1. Pass pointers to correspondent system DMA buffers.
	 2. Check and renew CAS signalization.
	 3. Check loosing of sync.
	 4. Perform echo cancelation on RX chanks.
	 5. Call dahdi_receive and dahdi_transmit. //process independently
	 */
	//check in which part of DMA buffer we are...

    if(!ptmc_dev->priv_data)
    {
	return -1;
    }
	//check DMA overflow counter;
        keep = (ptmc_dev->bar0[0x09] & 0xff000000) >> 24;
    	if (keep != pmc_zap->ovfl_counter)
	    module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", keep);
	pmc_zap->ovfl_counter = keep;

	keep = ptmc_dev->bar0[0x0b]; //here we get relative addres of current dma operating position.

	i = 8 + (keep/512)-(pmc_zap->keep/512)-1;
	if (i >= 8) {
            pmc_zap->irqmisses += (int)i-8;
	} else {
            pmc_zap->irqmisses += (int)i;
	}
	pmc_zap->keep = keep;

	if (keep > 3583) {
            //read offset 3072, from this address to read
            //write offset 512 (is it correct?), to this address to write
            SET_POINTERS(3072, 512);

        } else if (keep > 3071) {
              //2560
              //0
            SET_POINTERS(2560, 0);
        } else if (keep > 2559) {
             //2048
             //3584
            SET_POINTERS(2048, 3584);
        } else if (keep > 2047) {
            //1536
            //3072
            SET_POINTERS(1536, 3072);
        } else if (keep > 1535) {
              //1024
              //2560
            SET_POINTERS(1024, 2560);
        } else if (keep > 1023) {
              //512
              //2048
            SET_POINTERS(512, 2048);
        } else if (keep > 511) {
              //0
              //1536
            SET_POINTERS(0, 1536);
        } else { //keep > 0
              //start offset 3584
            //1024
            SET_POINTERS(3584, 1024);
        }

	//Check signalisation
	for (i=0; i < pmc_zap->num_spans; i++) {
		if(pmc_zap->e12zap_pmc_span[i]->span.flags & DAHDI_FLAG_RUNNING) {
			for(j1=1; j1 < 16; j1++) //chans < 16                            dahdi channels [1:15]  - chans [0:14]
			{
				tmp1 = (ptmc_dev->bar1[i * 0x40 + CAS_BIT_BUFFER_OFFSET + j1] >> 12 ) & 0x0f;
				if (tmp1 != pmc_zap->e12zap_pmc_span[i]->span.chans[j1-1]->rxsig)
				{
					if(!(pmc_zap->e12zap_pmc_span[i]->span.chans[j1-1]->sig & DAHDI_SIG_CLEAR))
					{
					  //      printk("Call, tmp1 =%x on ts:%d\n", tmp1, j1);
						dahdi_rbsbits(pmc_zap->e12zap_pmc_span[i]->span.chans[j1-1], tmp1);
					}
				}
			}
		
			for(j1=1;j1<16;j1++) //chans > 16                            dahdi channels [17:31] -chans [16:30]
			{
				tmp1 = (ptmc_dev->bar1[i * 0x40 + CAS_BIT_BUFFER_OFFSET + j1] >> 8) & 0x0f;
				if (tmp1 != pmc_zap->e12zap_pmc_span[i]->span.chans[15+j1]->rxsig)
				{
					if(!(pmc_zap->e12zap_pmc_span[i]->span.chans[15+j1]->sig & DAHDI_SIG_CLEAR))
					{
						  //      printk("Call, tmp1 =%x on ts:%d\n", tmp1, j1+15);
						dahdi_rbsbits(pmc_zap->e12zap_pmc_span[i]->span.chans[15+j1], tmp1);
					}
				}
			}
		}
	} 

/*This section can be used to pass fake CAS using PTMC_E1_PASS_ABCD_ZAP_FAKE ioctl.
	//Put fake signalisations
#define GET_CHAN_(a) ((a >> 4) & 0xFFFFF)
#define GET_SPAN_(a) (a >> 9)
	int_sig = ptmc_dev->test_sig;
	if ((GET_SPAN_(int_sig) < pmc_zap->num_spans) && (GET_CHAN_(int_sig) < 30))
	{
	    if ((int_sig & 0xf) != pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)].rxsig)
	    {
		if(!(pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)].sig & DAHDI_SIG_CLEAR))
		{
		    module_dprintk(KERN_DEBUG, "Passing sig: span: %d channel:%d, abcd=%01x\n", GET_SPAN_(int_sig)+1, GET_CHAN_(int_sig)+1, (int_sig & 0xf));
		    dahdi_rbsbits(&pmc_zap->e12zap_pmc_span[GET_SPAN_(int_sig)]->span.chans[GET_CHAN_(int_sig)], int_sig & 0xf);
		}
	    }
	 }
*/
	e12zap_proc(pmc_zap);
	++pmc_zap->intcount;;
        if (0 == (pmc_zap->intcount & 127)) {
	    e12zap_check_alarm (pmc_zap);
	    e12zap_find_master (pmc_zap);
            e12zap_clearstat (pmc_zap);
	}
	return 0;
}


int e12zap_dacs(struct dahdi_chan *dst, struct dahdi_chan *src)
{
	struct e12zap_span * e12zap_pmc_span_dst = dst->pvt;
	struct e12zap_pmc * pmc_zap = e12zap_pmc_span_dst->pvt;
	struct pmc_e1_controller * tmp_dev = pmc_zap->pvt;

	if (!dst) return -1;
	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");
	if(src)
	{
		struct e12zap_span * e12zap_pmc_span_src = src->pvt;
		if(e12zap_pmc_span_dst->pvt != e12zap_pmc_span_src->pvt) //trying to cross-connect timeslots from different boards
		{
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");
			//here to return to default crossing
			//TO E1 ts												 //FROM DMA ts, 2 - is DMA buffer offset
			tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_dst->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE((2 + e12zap_pmc_span_dst->spannumber), dst->chanpos);
			tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_src->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE((2 + e12zap_pmc_span_src->spannumber), src->chanpos);
	
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", e12zap_pmc_span_dst->spannumber, dst->chanpos, e12zap_pmc_span_src->spannumber, src->chanpos);
		//TO E1 ts         											 //FROM E1 ts
		tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_dst->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE(e12zap_pmc_span_src->spannumber, src->chanpos);
		tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_src->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE(e12zap_pmc_span_dst->spannumber, dst->chanpos);
		return 0;			
	}
	else
	{
	    //return to default crossing
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", e12zap_pmc_span_dst->spannumber, dst->chanpos);
		//TO E1 ts												 //FROM DMA ts, 2 - is DMA buffer offset
		tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_dst->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE((2 + e12zap_pmc_span_dst->spannumber), dst->chanpos);
		//tmp_dev->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * e12zap_pmc_span_src->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE(2 + e12zap_pmc_span_src->spannumber, src->chanpos);
		
		return 0;
	}
}
